==> oobium_datastore-1.0.0-dev.6.tar.gz: lib/src/state.dart <==
import 'dart:async';

import 'package:oobium_datastore/src/executor.dart';
import 'package:oobium_logging/oobium_logging.dart';

abstract class StateKey<T> {
  dynamic get id;
  const StateKey();
  @override
  bool operator ==(Object other) => identical(this, other)
      || (other is StateKey<T> && id == other.id);
  @override
  int get hashCode => Object.hash(T, id);
  @override
  String toString() => '$runtimeType: $T($id)';
}

abstract class StateValue<T> extends StateKey<T> implements Comparable<T> {
  const StateValue();
}

class StateStore {
  final _log = Logger('oobium_datastore').child(StateStore);
  final _state = <StateKey, StateValue>{};

  void Function(StateEvent e)? _onEvent;

  Future<void> clear() async {
    _state.clear();
  }

  Future<void> close() async {
    _onEvent = null;
    _state.clear();
  }

  int get length => _state.length;
  bool get isEmpty => _state.isEmpty;
  bool get isNotEmpty => _state.isNotEmpty;

  bool contains(Object? obj) => (obj is StateKey) && _state.containsKey(obj);

  T get<T extends StateValue>(StateKey key) {
    final state = _state[key];
    if(state is T) return state;
    if(state == null) throw Exception('not found: $key');
    throw Exception('$state is not of type $T');
  }

  T? getOrNull<T extends StateValue>(StateKey key) {
    final state = _state[key];
    if(state is T?) return state;
    throw Exception('$state is not of type $T');
  }

  Iterable<T> getAll<T extends StateValue>(bool Function(T s)? test) {
    return (test != null)
      ? _state.values.whereType<T>().where(test).toList()
      : _state.values.whereType<T>();
  }

  Iterable<StateValue> getWhere(bool Function(StateValue s) test) {
    return _state.values.where(test);
  }

  void put<T extends StateValue>(StateFeature source, T value) {
    final old = _state[value];
    if(old?.compareTo(value) != 0) {
      final e = StateEvent<T>._(source, old, value);
      _state[value] = value;
      _onEvent?.call(e);
      _log.debug(() => e.isCreate ? '+created<$T>($value)' : '+updated<$T>($value)');
    }
  }

  T? remove<T extends StateValue>(StateFeature source, StateKey<T>? key) {
    if(key != null) {
      final removed = _state.remove(key);
      if(removed is T) {
        final e = StateEvent<T>._remove(source, removed);
        _onEvent?.call(e);
        _log.debug(() => '+removed($key)');
        return removed;
      }
      if(removed != null) {
        _state[key] = removed;
        throw Exception('$removed is not of type $T; it has been restored');
      }
    }
    return null;
  }

  Future<void> reset(Stream<StateValue> values) async {
    _log.debug(() => 'resetting state...');
    clear();
    _log.debug(() => '  state cleared');
    await for(final value in values) {
      _state[value] = value;
      _log.log(debug: '  added value', trace: () => '  added $value');
    }
    _log.debug(() => '  reset complete.');
  }

  void set<T extends StateValue>(StateFeature source, Iterable<T> list, {bool Function(T s)? where}) {
    final set = list.toSet();
    getAll<T>((s) => !set.contains(s) && (where == null || where(s)))
        .forEach((s) => remove<T>(source, s as StateKey<T>));
    list.forEach((s) => put(source, s));
  }
}

class StateEvent<T extends StateValue> {
  final StateFeature source;
  final T? _oldValue;
  final T? _newValue;
  StateEvent._(this.source, StateValue? oldState, this._newValue) :
    assert(T != StateValue),
    assert(oldState != null || _newValue != null),
    assert(oldState == null || _newValue == null || oldState == _newValue),
    _oldValue = oldState as T?;
  StateEvent._remove(this.source, T oldState) :
    assert(T != StateValue),
    _oldValue = oldState,
    _newValue = null;

  bool get isCreate => _oldValue == null && _newValue != null;
  bool get isNotCreate => !isCreate;
  bool get isUpdate => _oldValue != null && _newValue != null;
  bool get isNotUpdate => !isUpdate;
  bool get isRemove => _oldValue != null && _newValue == null;
  bool get isNotRemove => !isRemove;

  T get value => _newValue ?? _oldValue!;

  T get oldValue => _oldValue ?? (throw Exception('Create event does not have an oldValue'));
  T get newValue => _newValue ?? (throw Exception('Remove event does not have an newValue'));

  @override
  String toString() => '$source<$T>($_oldValue -> $_newValue)';

  StateEvent<T> _toCreate() => StateEvent<T>._(source, null, _newValue);
}

class StateEvents {
  final _events = <StateValue, StateEvent>{};
  final _ripples = <StateKey>{};
  StateEvents._();

  List<StateEvent> created(StateFeature from) => _events.values.where((e) => e.source != from && e.isCreate).toList();
  List<StateEvent> updated(StateFeature from) => _events.values.where((e) => e.source != from && e.isUpdate).toList();
  List<StateEvent> removed(StateFeature from) => _events.values.where((e) => e.source != from && e.isRemove).toList();

  void _addEvent(StateEvent e) {
    if(e.isCreate) {
      _events[e.value] = e;
    }
    else if(e.isUpdate) {
      if(_events[e.value]?.isCreate == true) {
        // create/update in one cycle: update the create event
        _events[e.value] = e._toCreate();
      } else {
        _events[e.value] = e;
      }
    }
    else if(e.isRemove) {
      if(_events.remove(e.value)?.isCreate == true) {
        // create/remove in one cycle: remove the create event
      } else {
        _events[e.value] = e;
      }
    }
  }

  bool _addRipple(StateKey k) {
    return !_events.containsKey(k) && _ripples.add(k);
  }

  bool contains(StateKey? key) => (key != null)
      && (_events.containsKey(key) || _ripples.contains(key));

  int get length => _events.length;
  bool get isEmpty => _events.isEmpty;
  bool get isNotEmpty => !isEmpty;

  Iterable<StateValue> get values => _events.values.map((e) => e.value);

  @override
  String toString() {
    final created = _events.values.where((e) => e.isCreate);
    final updated = _events.values.where((e) => e.isUpdate);
    final removed = _events.values.where((e) => e.isRemove);
    return '$runtimeType(${[
      'created: [${created.join(', ')}]',
      'updated: [${updated.join(', ')}]',
      'removed: [${removed.join(', ')}]',
    ].join(', ')})';
  }
}

class StateReader {
  final StateFeature _feature;
  StateReader._(this._feature);
  bool contains<T extends StateValue<T>>(StateKey<T>? key) => _feature.contains(key);
  T get<T extends StateValue<T>>(StateKey<T> key) => _feature.get(key);
}

class StateWriter {
  final StateFeature _feature;
  StateWriter._(this._feature);
  bool contains<T extends StateValue>(StateKey<T>? key) => _feature.contains(key);
  T get<T extends StateValue>(StateKey<T> key) => _feature.get(key);
  void put<T extends StateValue>(T state) => _feature.put(state);
  T? remove<T extends StateValue>(StateKey<T>? key) => (key != null) ? _feature.remove(key) : null;
}

class StateContext {

  final StateStore _state;
  final _features = <StateFeature>[];
  Executor? _executor;
  StateContext({Iterable<StateFeature>? features}) :
    _state = StateStore()
  {
    if(features != null) {
      for(final feature in features) {
        feature._context = this;
      }
      _features.addAll(features);
    }
  }

  T feature<T extends StateFeature>() => _features.whereType<T>().first;

  Executor get _q => (_executor ??= Executor());

  var _open = false;
  Future<void> open() async {
    if(isNotOpen) {
      _open = true;
      await Future.wait(_features.map((f) => f.onOpen()));
      _state._onEvent = _onEvent;
    }
  }

  Future<void> flush() async {
    await Future.wait(_features.map((f) => f.onFlush()));
    await Future.value();
    await _q.flush();
  }

  Future<void> close() async {
    _open = false;
    await Future.wait(_features.map((f) => f.onClose()));
    await Future.value();
    await _q.flush();
    await _state.close();
  }

  Future<void> destroy() async {
    _open = false;
    await _state.close();
    await Future.wait(_features.map((f) => f.onDestroy()));
    await _q.cancel();
    _executor = null;
  }

  Future<void> reset([Stream<StateValue>? values]) async {
    await destroy();
    if(values != null) {
      await _state.reset(values);
    }
    await open();
  }

  void _onEvent(StateEvent event) {
    for(final feature in _features) {
      feature.onEvent(event);
    }
  }

  bool contains(Object? key) => _state.contains(key);
  T get<T extends StateValue>(StateKey<T> key) => _state.get(key);
  T? getOrNull<T extends StateValue>(StateKey<T> key) => _state.getOrNull(key);
  Iterable<T> getAll<T extends StateValue<T>>(bool Function(T s)? test) => _state.getAll<T>(test);
  Iterable<StateValue> getWhere(bool Function(StateValue s) test) => _state.getWhere(test);

  void _put<T extends StateValue>(StateFeature source, T state) => _state.put(source, state);
  T? _remove<T extends StateValue>(StateFeature source, StateKey<T>? key) => (key != null) ? _state.remove(source, key) : null;
  Future<void> _reset(StateFeature source, Stream<StateValue> values) => _state.reset(values);
  void _set<T extends StateValue>(StateFeature source, Iterable<T> list, {bool Function(T s)? where}) => _state.set(source, list, where: where);

  Iterable<StateKey> get stateKeys=> _state._state.keys;
  Iterable<StateValue> get stateValues=> _state._state.values;

  int get length => _state.length;
  bool get isEmpty => _state.isEmpty;
  bool get isNotEmpty => _state.isNotEmpty;
  bool get isOpen => _open;
  bool get isNotOpen => !isOpen;
}

abstract class StateFeature {
  late final _log = Logger(this);
  late final StateContext _context;

  T call<T extends StateFeature>() => _context.feature<T>();
  StateReader asReader() => StateReader._(this);
  StateWriter asWriter() => StateWriter._(this);

  Executor get q => _context._q;

  int get length => _context.length;
  Iterable<StateKey> get stateKeys => _context.stateKeys;
  Iterable<StateValue> get stateValues => _context.stateValues;

  bool get isOpen => _context.isOpen;
  bool get isNotOpen => _context.isNotOpen;

  bool contains(Object? key) => _context.contains(key);
  T get<T extends StateValue>(StateKey<T> key) => _context.get(key);
  T? getOrNull<T extends StateValue>(StateKey<T> key) => _context.getOrNull(key);
  Iterable<T> getAll<T extends StateValue<T>>(bool Function(T s)? test) => _context.getAll<T>(test);
  Iterable<StateValue> getWhere(bool Function(StateValue s) test) => _context.getWhere(test);

  void put<T extends StateValue>(T state) => _context._put(this, state);
  T? remove<T extends StateValue>(StateKey<T>? key) => (key != null) ? _context._remove(this, key) : null;
  Future<void> reset(Stream<StateValue> values) => _context._reset(this, values);
  void set<T extends StateValue>(Iterable<T> list, {bool Function(T s)? where}) => _context._set(this, list, where: where);

  Future<void> onOpen() {
    _log.trace(() => 'onOpen()');
    return Future.value();
  }
  Future<void> onFlush() {
    _log.trace(() => 'onFlush()');
    return Future.value();
  }
  Future<void> onClose() {
    _log.trace(() => 'onClose()');
    return Future.value();
  }
  Future<void> onDestroy() {
    _log.trace(() => 'onDestroy()');
    return Future.value();
  }
  void onEvent(StateEvent event) {
    _log.trace(() => 'onEvent($event)');
  }
}

class BufferedStateFeature extends StateFeature {
  StateEvents? _events;

  void flushEvents() {
    final e = _events;
    if(e != null && e.isNotEmpty) {
      onEvents(e);
    }
    _events = null;
  }

  // TODO other lifecycle events

  @override
  void onEvent(StateEvent event) {
    super.onEvent(event);
    if(_events == null) {
      scheduleMicrotask(flushEvents);
    }
    (_events ??= StateEvents._())._addEvent(event);
  }

  bool addRipple(StateKey key) {
    return _events?._addRipple(key) ?? false;
  }

  void onEvents(StateEvents events) {
    _log.trace(() => 'onEvents($events)');
  }
}

abstract class WatchableStateFeature extends BufferedStateFeature {

  StreamController<StateEvents>? _controller;
  StreamController<StateEvents> _newController() {
    return StreamController<StateEvents>.broadcast(
        onCancel: () => _controller = null
    );
  }

  Stream<StateEvents> watch() => (_controller ??= _newController()).stream;

  @override
  Future<void> onClose() async {
    await _controller?.close();
    _controller = null;
    return super.onClose();
  }

  @override
  Future<void> onDestroy() async {
    await _controller?.close();
    _controller = null;
    return super.onClose();
  }

  @override
  void onEvents(StateEvents events) {
    final c = _controller;
    if(c != null && events.isNotEmpty) {
      c.add(events);
    }
  }
}

==> oobium_datastore-1.1.1.tar.gz: lib/src/state.dart <==
import 'dart:async';

import 'package:objectid/objectid.dart';
import 'package:oobium_datastore/src/executor.dart';
import 'package:oobium_logging/oobium_logging.dart' show Logger;

abstract class StateKey<T> {
  dynamic get id;
  const StateKey();
  @override
  bool operator ==(Object other) => identical(this, other)
      || (other is StateKey<T> && id == other.id);
  @override
  int get hashCode => Object.hash(T, id);
  @override
  String toString() => '$runtimeType: $T($id)';
}

abstract class StateValue<T> extends StateKey<T> implements Comparable<T> {
  ObjectId? _updateId;
  StateValue(this._updateId);
  ObjectId get updateId => _updateId ?? (throw Exception('updateId not set'));
  ObjectId _touch() => _updateId ??= ObjectId();
}

class StateStore {
  final _log = Logger('oobium_datastore').child(StateStore);
  final _state = <StateKey, StateValue>{};

  var _lastUpdateId = ObjectId.fromTimestamp(DateTime(2021, 4, 4));
  void Function(StateEvent e)? _onEvent;

  Future<void> clear() async {
    _state.clear();
  }

  Future<void> close() async {
    _onEvent = null;
    _state.clear();
  }

  int get length => _state.length;
  bool get isEmpty => _state.isEmpty;
  bool get isNotEmpty => _state.isNotEmpty;

  bool contains(Object? obj) => (obj is StateKey) && _state.containsKey(obj);

  T get<T extends StateValue>(StateKey key) {
    final state = _state[key];
    if(state is T) return state;
    if(state == null) throw Exception('not found: $key');
    throw Exception('$state is not of type $T');
  }

  T? getOrNull<T extends StateValue>(StateKey key) {
    final state = _state[key];
    if(state is T?) return state;
    throw Exception('$state is not of type $T');
  }

  Iterable<T> getAll<T extends StateValue>(bool Function(T s)? test) {
    return (test != null)
      ? _state.values.whereType<T>().where(test).toList()
      : _state.values.whereType<T>();
  }

  Iterable<StateValue> getWhere(bool Function(StateValue s) test) {
    return _state.values.where(test);
  }

  void put<T extends StateValue>(StateFeature source, T value) {
    final old = _state[value];
    if(old?.compareTo(value) != 0) {
      _lastUpdateId = value._touch();
      _state[value] = value;
      final e = StateEvent<T>._(source, old, value);
      _onEvent?.call(e);
      _log.debug(() => e.isCreate ? '+created<$T>($value)' : '+updated<$T>($value)');
    }
  }

  T? remove<T extends StateValue>(StateFeature source, StateKey<T>? key) {
    if(key != null) {
      final removed = _state.remove(key);
      if(removed is T) {
        _lastUpdateId = removed._touch();
        final e = StateEvent<T>._remove(source, removed);
        _onEvent?.call(e);
        _log.debug(() => '+removed($key)');
        return removed;
      }
      if(removed != null) {
        _state[key] = removed;
        throw Exception('$removed is not of type $T; it has been restored');
      }
    }
    return null;
  }

  Future<void> reset(Stream<StateValue> values) async {
    _log.debug(() => 'resetting state...');
    clear();
    _log.debug(() => '  state cleared');
    await for(final value in values) {
      _state[value] = value;
      _log.log(debug: '  added value', trace: () => '  added $value');
    }
    _log.debug(() => '  reset complete.');
  }

  void set<T extends StateValue>(StateFeature source, Iterable<T> list, {bool Function(T s)? where}) {
    final set = list.toSet();
    getAll<T>((s) => !set.contains(s) && (where == null || where(s)))
        .forEach((s) => remove<T>(source, s as StateKey<T>));
    list.forEach((s) => put(source, s));
  }
}

class StateEvent<T extends StateValue> {
  final StateFeature source;
  final T? _oldValue;
  final T? _newValue;
  StateEvent._(this.source, StateValue? oldState, this._newValue) :
    assert(T != StateValue),
    assert(oldState != null || _newValue != null),
    assert(oldState == null || _newValue == null || oldState == _newValue),
    _oldValue = oldState as T?;
  StateEvent._remove(this.source, T oldState) :
    assert(T != StateValue),
    _oldValue = oldState,
    _newValue = null;

  bool get isCreate => _oldValue == null && _newValue != null;
  bool get isNotCreate => !isCreate;
  bool get isUpdate => _oldValue != null && _newValue != null;
  bool get isNotUpdate => !isUpdate;
  bool get isRemove => _oldValue != null && _newValue == null;
  bool get isNotRemove => !isRemove;

  T get value => _newValue ?? _oldValue!;

  T get oldValue => _oldValue ?? (throw Exception('Create event does not have an oldValue'));
  T get newValue => _newValue ?? (throw Exception('Remove event does not have an newValue'));

  @override
  String toString() => '$source<$T>($_oldValue -> $_newValue)';

  StateEvent<T> _toCreate() => StateEvent<T>._(source, null, _newValue);
}

class StateEvents {
  final _events = <StateValue, StateEvent>{};
  final _ripples = <StateKey>{};
  StateEvents._();

  List<StateEvent> created(StateFeature from) => _events.values.where((e) => e.source != from && e.isCreate).toList();
  List<StateEvent> updated(StateFeature from) => _events.values.where((e) => e.source != from && e.isUpdate).toList();
  List<StateEvent> removed(StateFeature from) => _events.values.where((e) => e.source != from && e.isRemove).toList();

  void _addEvent(StateEvent e) {
    if(e.isCreate) {
      _events[e.value] = e;
    }
    else if(e.isUpdate) {
      if(_events[e.value]?.isCreate == true) {
        // create/update in one cycle: update the create event
        _events[e.value] = e._toCreate();
      } else {
        _events[e.value] = e;
      }
    }
    else if(e.isRemove) {
      if(_events.remove(e.value)?.isCreate == true) {
        // create/remove in one cycle: remove the create event
      } else {
        _events[e.value] = e;
      }
    }
  }

  bool _addRipple(StateKey k) {
    return !_events.containsKey(k) && _ripples.add(k);
  }

  bool contains(StateKey? key) => (key != null)
      && (_events.containsKey(key) || _ripples.contains(key));

  int get length => _events.length;
  bool get isEmpty => _events.isEmpty;
  bool get isNotEmpty => !isEmpty;

  Iterable<StateValue> get values => _events.values.map((e) => e.value);

  @override
  String toString() {
    final created = _events.values.where((e) => e.isCreate);
    final updated = _events.values.where((e) => e.isUpdate);
    final removed = _events.values.where((e) => e.isRemove);
    return '$runtimeType(${[
      'created: [${created.join(', ')}]',
      'updated: [${updated.join(', ')}]',
      'removed: [${removed.join(', ')}]',
    ].join(', ')})';
  }
}

class StateContext {
  late final _log = Logger(this);
  final StateStore _state;
  final StateIdProvider? _idProvider;
  final _features = <StateFeature>[];
  Executor? _executor;
  StateContext({StateIdProvider? idProvider, Iterable<StateFeature>? features}) :
    _state = StateStore(),
    _idProvider = idProvider
  {
    if(features != null) {
      for(final feature in features) {
        feature._context = this;
      }
      _features.addAll(features);
    }
  }

  T feature<T extends StateFeature>() => _features.whereType<T>().first;

  Executor get _q => (_executor ??= Executor());

  String? _id;
  String get id => (_id ?? ObjectId().hexString);

  var _open = false;
  Future<void> open() async {
    if(isNotOpen) {
      _open = true;
      _id = _idProvider?.id;
      await Future.wait(_features.map(_onOpen));
      _state._onEvent = _onEvent;
    }
  }

  Future<void> flush() async {
    await Future.wait(_features.map(_onFlush));
    await Future.value();
    await _q.flush();
  }

  Future<void> close() async {
    _open = false;
    await Future.wait(_features.map(_onClose));
    await Future.value();
    await _q.flush();
    await _state.close();
  }

  Future<void> destroy() async {
    _open = false;
    await _state.close();
    await Future.wait(_features.map(_onDestroy));
    await _q.cancel();
    _executor = null;
  }

  Future<void> reset([Stream<StateValue>? values]) async {
    await destroy();
    if(values != null) {
      await _state.reset(values);
    }
    await open();
  }

  Future<void> _onOpen(StateFeature feature) {
    _log.trace(() => 'onOpen($feature)');
    return feature.onOpen();
  }

  Future<void> _onFlush(StateFeature feature) {
    _log.trace(() => 'onFlush($feature)');
    return feature.onFlush();
  }

  Future<void> _onClose(StateFeature feature) {
    _log.trace(() => 'onClose($feature)');
    return feature.onClose();
  }

  Future<void> _onDestroy(StateFeature feature) {
    _log.trace(() => 'onDestroy($feature)');
    return feature.onDestroy();
  }

  void _onEvent(StateEvent event) {
    for(final feature in _features) {
      _log.trace(() => 'onEvent($feature, $event)');
      feature.onEvent(event);
    }
  }

  bool contains(Object? key) => _state.contains(key);
  T get<T extends StateValue>(StateKey<T> key) => _state.get(key);
  T? getOrNull<T extends StateValue>(StateKey<T> key) => _state.getOrNull(key);
  Iterable<T> getAll<T extends StateValue<T>>(bool Function(T s)? test) => _state.getAll<T>(test);
  Iterable<StateValue> getWhere(bool Function(StateValue s) test) => _state.getWhere(test);

  void _put<T extends StateValue>(StateFeature source, T state) => _state.put(source, state);
  T? _remove<T extends StateValue>(StateFeature source, StateKey<T>? key) => (key != null) ? _state.remove(source, key) : null;
  Future<void> _reset(StateFeature source, Stream<StateValue> values) => _state.reset(values);
  void _set<T extends StateValue>(StateFeature source, Iterable<T> list, {bool Function(T s)? where}) => _state.set(source, list, where: where);

  Iterable<StateKey> get stateKeys=> _state._state.keys;
  Iterable<StateValue> get stateValues=> _state._state.values;

  int get length => _state.length;
  bool get isEmpty => _state.isEmpty;
  bool get isNotEmpty => _state.isNotEmpty;
  bool get isOpen => _open;
  bool get isNotOpen => !isOpen;

  @override
  String toString() => '$runtimeType';
}

abstract class StateIdProvider {
  String get id;
}

abstract class StateFeature {
  late final log = Logger(this);
  late final StateContext _context;

  T call<T extends StateFeature>() => _context.feature<T>();

  Executor get q => _context._q;

  ObjectId get lastUpdateId => _context._state._lastUpdateId;
  int get length => _context.length;
  Iterable<StateKey> get stateKeys => _context.stateKeys;
  Iterable<StateValue> get stateValues => _context.stateValues;

  bool get isOpen => _context.isOpen;
  bool get isNotOpen => _context.isNotOpen;

  bool contains(Object? key) => _context.contains(key);
  T get<T extends StateValue>(StateKey<T> key) => _context.get(key);
  T? getOrNull<T extends StateValue>(StateKey<T> key) => _context.getOrNull(key);
  Iterable<T> getAll<T extends StateValue<T>>(bool Function(T s)? test) => _context.getAll<T>(test);
  Iterable<StateValue> getWhere(bool Function(StateValue s) test) => _context.getWhere(test);

  void put<T extends StateValue>(T state) => _context._put(this, state);
  T? remove<T extends StateValue>(StateKey<T>? key) => (key != null) ? _context._remove(this, key) : null;
  Future<void> reset(Stream<StateValue> values) => _context._reset(this, values);
  void set<T extends StateValue>(Iterable<T> list, {bool Function(T s)? where}) => _context._set(this, list, where: where);

  Future<void> onOpen() => Future.value();
  Future<void> onFlush() => Future.value();
  Future<void> onClose() => Future.value();
  Future<void> onDestroy() => Future.value();
  void onEvent(StateEvent event) { }

  @override
  String toString() => '$runtimeType';
}

class BufferedStateFeature extends StateFeature {
  StateEvents? _events;

  void flushEvents() {
    final e = _events;
    if(e != null && e.isNotEmpty) {
      onEvents(e);
    }
    _events = null;
  }

  // TODO other lifecycle events

  @override
  void onEvent(StateEvent event) {
    super.onEvent(event);
    if(_events == null) {
      scheduleMicrotask(flushEvents);
    }
    (_events ??= StateEvents._())._addEvent(event);
  }

  bool addRipple(StateKey key) {
    return _events?._addRipple(key) ?? false;
  }

  void onEvents(StateEvents events) { }
}

abstract class WatchableStateFeature extends BufferedStateFeature {

  StreamController<StateEvents>? _controller;
  StreamController<StateEvents> _newController() {
    return StreamController<StateEvents>.broadcast(
        onCancel: () => _controller = null
    );
  }

  Stream<StateEvents> watch() => (_controller ??= _newController()).stream;

  @override
  Future<void> onClose() async {
    await _controller?.close();
    _controller = null;
    return super.onClose();
  }

  @override
  Future<void> onDestroy() async {
    await _controller?.close();
    _controller = null;
    return super.onClose();
  }

  @override
  void onEvents(StateEvents events) {
    final c = _controller;
    if(c != null && events.isNotEmpty) {
      c.add(events);
    }
  }
}
